unit QuatrisCtrl;

interface

uses
  Classes;

type
  TKleur = Byte;
  TKolom = 0..9;
  TRij = 0..19;
  TBord = array[0..19, 0..9] of TKleur;
  TVeld = record
    Kolom: TKolom;
    Rij: TRij;
  end;

  TBlok = array[0..3] of TVeld;

  TQuatris = class(TObject)
  private
    FBord: TBord;
    FCurrentBlok: TBlok;
    FCanFall: Boolean;
    FBlokKleur: TKleur;
    FGameOver: Boolean;
    FScore: Integer;
    FOnSetCurrentBlok: TNotifyEvent;
    FOnFallBlock: TNotifyEvent;
    FOnScoreChange: TNotifyEvent;
    FOnGameOver: TNotifyEvent;
    FOnNewGame: TNotifyEvent;
    function GetBord(Rij: TRij; Kolom: TKolom): TKleur;
    procedure NewBlok;
    function CanBlok(Blok: TBlok): Boolean;
    procedure GameOver;
    procedure SetCurrentBlok(ABlok: TBlok);
    procedure SetScore(Value: Integer);
  public
    constructor Create; virtual;
    procedure NewGame;
    procedure BlokToRight;
    procedure BlokToLeft;
    procedure FallBlok;
    procedure Rotate;
    property CurrentBlok: TBlok read FCurrentBlok;
    property BlokKleur: TKleur read FBlokKleur;
    property Bord[Rij: TRij; Kolom: TKolom]: TKleur read GetBord;
    property Score: Integer read FScore write SetScore;
    property AGameOver: Boolean read FGameOver;
    property CanFall: Boolean read FCanFall;
    property OnSetCurrentBlok: TNotifyEvent read FOnSetCurrentBlok write FOnSetCurrentBlok;
    property OnFallBlock: TNotifyEvent read FOnFallBlock write FOnFallBlock;
    property OnScoreChange: TNotifyEvent read FOnScoreChange write FOnScoreChange;
    property OnGameOver: TNotifyEvent read FOnGameOver write FOnGameOver;
    property OnNewGame: TNotifyEvent read FOnNewGame write FOnNewGame;
  end;

implementation

// Range checking is handled insite the code.
{$R-}

constructor TQuatris.Create;
begin
  inherited Create;
  NewGame;
end;

function TQuatris.GetBord(Rij: TRij; Kolom: TKolom): TKleur;
begin
  Result := FBord[Rij, Kolom];
end;

procedure TQuatris.NewGame;
begin
  FGameOver := False;
  FillChar(FBord[0,0], 10 * 20 * SizeOf(TKleur), 0);
  Score := 0;
  NewBlok;
  if Assigned(OnNewGame) then
    OnNewGame(Self);
end;

procedure TQuatris.SetCurrentBlok(ABlok: TBlok);
var
  I: Integer;
begin
  FCurrentBlok := ABlok;
  // Check if the block can fall.
  for I := 0 to 3 do
    Inc(ABlok[I].Rij);
  FCanFall := CanBlok(ABlok);
  // Call the event handler
  if Assigned(OnSetCurrentBlok) then
    OnSetCurrentBlok(Self);
end;

procedure TQuatris.SetScore(Value: Integer);
begin
  FScore := Value;
  if Assigned(OnScoreChange) then
    OnScoreChange(Self);
end;

procedure TQuatris.NewBlok;
const
  Bloks: array[0..6] of TBlok =
  (((Kolom: 4; Rij: 0), //  ##
    (Kolom: 4; Rij: 1), //  ##
    (Kolom: 5; Rij: 0),
    (Kolom: 5; Rij: 1)),
   ((Kolom: 4; Rij: 0), // ###
    (Kolom: 3; Rij: 0), //  #
    (Kolom: 5; Rij: 0),
    (Kolom: 4; Rij: 1)),
   ((Kolom: 4; Rij: 0), // ###
    (Kolom: 3; Rij: 0), // #
    (Kolom: 5; Rij: 0),
    (Kolom: 3; Rij: 1)),
   ((Kolom: 4; Rij: 0), // ###
    (Kolom: 3; Rij: 0),  //   #
    (Kolom: 5; Rij: 0),
    (Kolom: 5; Rij: 1)),
   ((Kolom: 4; Rij: 0),  //  ##
    (Kolom: 5; Rij: 0),  // ##
    (Kolom: 3; Rij: 1),
    (Kolom: 4; Rij: 1)),
   ((Kolom: 4; Rij: 0), // ##
    (Kolom: 3; Rij: 0),  //  ##
    (Kolom: 4; Rij: 1),
    (Kolom: 5; Rij: 1)),
   ((Kolom: 4; Rij: 0), // ####
    (Kolom: 3; Rij: 0),
    (Kolom: 5; Rij: 0),
    (Kolom: 6; Rij: 0)));
var
  ABlok: TBlok;
  I: Integer;
begin
  if FGameOver then Exit;
  // Makes a new blok on the Top
  I := round(Random(7));
  ABlok := Bloks[I];
  if not CanBlok(ABlok) then
    GameOver
  else
  begin
    FBlokKleur := I + 1;
    SetCurrentBlok(ABlok);
  end;
end;

function TQuatris.CanBlok(Blok: TBlok): Boolean;
var
  I: Integer;
begin
  // Test a blok can be on that position
  Result := True;
  for I := 0 to 3 do
  with Blok[I] do
  begin
    if (Kolom < 0) or (Kolom > 9) or (Rij < 0) or (Rij > 19) or
       (FBord[Rij, Kolom] <> 0) then
    begin
      Result := False;
      Exit;
    end;
  end;
end;

procedure TQuatris.GameOver;
begin
  FGameOver := True;
  if Assigned(FOnGameOver) then
    FOnGameOver(Self);
end;

procedure TQuatris.FallBlok;
var
  ABlok: TBlok;
  I: Integer;
  ARij: Integer;
  RemoveRijCount: Integer;
label
  NietVol;
begin
  if FGameOver then Exit;
  ABlok := FCurrentBlok;
  for I := 0 to 3 do
    Inc(ABlok[I].Rij);
  if FCanFall then
    SetCurrentBlok(ABlok)
  else
  begin
    // Blok kan niet meer vallen

    // Vul de plaatsen in het bord in
    for I := 0 to 3 do
    begin
      with FCurrentBlok[I] do
        FBord[Rij, Kolom] := FBlokKleur;
    end;
    // verwijder rijen
    RemoveRijCount := 0;
    for ARij := FCurrentBlok[0].Rij - 2 to FCurrentBlok[0].Rij + 2 do
    begin
      if (ARij < 0) or (ARij > 19) then
        goto NietVol;
      for I := 0 to 9 do
        if FBord[ARij, I] = 0 then
          goto NietVol;
      // Verwijder ARij
      move(FBord[0, 0], FBord[1, 0], SizeOf(TKleur) * 10 * ARij);
      FillChar(FBord[0, 0], SizeOf(TKleur) * 10, 0);
      Inc(RemoveRijCount);
NietVol:
    end;
    // Add punten voor nieuwe punten voor rij.
    if RemoveRijCount <> 0 then
    begin
      case RemoveRijCount of
        1: Inc(FScore, 1);
        2: Inc(FScore, 3);
        3: Inc(FScore, 5);
        4: Inc(FScore, 10);
      end;
      if Assigned(OnScoreChange) then
        OnScoreChange(Self);
    end;
    if Assigned(OnFallBlock) then
      OnFallBlock(Self);
    // Laad een nieuw blok
    NewBlok;
  end;
end;

procedure TQuatris.BlokToRight;
var
  ABlok: TBlok;
  I: Integer;
begin
  if FGameOver then Exit;
  ABlok := FCurrentBlok;
  for I := 0 to 3 do
    inc(ABlok[I].Kolom);
  if CanBlok(ABlok) then
    SetCurrentBlok(ABlok);
end;

procedure TQuatris.BlokToLeft;
var
  ABlok: TBlok;
  I: Integer;
begin
  if FGameOver then Exit;
  ABlok := FCurrentBlok;
  for I := 0 to 3 do
    dec(ABlok[I].Kolom);
  if CanBlok(ABlok) then
    SetCurrentBlok(ABlok);
end;

procedure TQuatris.Rotate;
var
  ARij: TRij;
  AKolom: TKolom;
  ATemp: Integer;
  ABlok: TBlok;
  I: Integer;
begin
  if FGameOver then Exit;
  ABlok := FCurrentBlok;
  ARij := ABlok[0].Rij;
  AKolom := ABlok[0].Kolom;
  for I := 1 to 3 do
    with ABlok[I] do
    begin
      ATemp := AKolom - Kolom;
      Kolom := AKolom + Rij -ARij;
      Rij := ARij + ATemp;
    end;
  if CanBlok(ABlok) then
    SetCurrentBlok(ABlok);
end;

end.
